/*********************************************************************

    Copyright 2016, Michael Steinert

    This file is part of DrawArea (Dijkstra Routing Algorithm Which
    Ascertains REachable Areas).

    DrawArea is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DrawArea is distributed WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DrawArea.  If not, see <http://www.gnu.org/licenses/>.

***********************************************************************/

/*************************************************

    TODO-List:
	- implement multi-threading

**************************************************/

#ifdef DRAWAREA_MPI
	#include "mpi.h"
	#define MPI_MSG(msg) "Proc %d " msg, ID
	#define MPI_ROOT_PROC 0
#else
	#define MPI_MSG(msg) msg
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string.h>

#include "DrawArea.h"
#include "ParseArgs.h"
#include "Error.h"
#include "pnm.h"

using namespace std;

int ID;

Node* ReadScopeFromMatrix(const char* fileName, Scope& scope);
int SplitHeuristic(const char* inputName, Scope inputScope, Scope* scopeList, size_t nColumns, size_t nRows, float limitTime, float pixelSize);
int CalculateAreaMap(const char* inputName, const char* outputName, Scope& scope, float limitTime, float minTime, float areaScaling, int nBinning);
Scope& Fit2Binning(Scope& scope, int nBinning, int xOffset, int yOffset);
size_t* Split2EqualParts(size_t nValues, double* values, size_t nParts, size_t *parts=NULL);
float MeanTravelTime(size_t *origin, size_t dataWidth, float pixelSize, int& validNeighbors);

double PixelWeight(double time, double limit)
{
	return pow(max(1.0-time/limit,0.0),4);
}

char helpText[]="Usage: %s [OPTIONS] -l TIME -p PIXELSIZE input\n"
		"\nMandatories:\n"
		"  -l, --limit TIME\n"
			"\ttravel time limit in seconds\n"
		"  -p, --pixel-size PIXELSIZE\n"
			"\tinput image pixel size in meters\n"
		"\nOptions:\n"
		"  -h, --help\n"
			"\tprint this message\n"
		"  -a, --area AREA\n"
			"\toutput data will be scaled to AREA square kilometers\n"
		"  -b, --binning N\n"
			"\tcalculate only every Nth point in each row and column\n"
		"  -o, --output FILE\n"
			"\tforce output to be written to FILE\n"
		"  -s, --scope WIDTHxHEIGHT+LEFT+TOP\n"
			"\tset scope of interest for map calculation\n"
#ifndef DRAWAREA_MPI
		"  -c, --cols N\n"
			"\tforce map splitting into [N] columns\n"
		"  -r, --rows N\n"
			"\tforce map splitting into N rows\n"
#endif
#ifdef DRAWAREA_THREADS
		"  -n, --no-calc\n"
			"\tdon't perform calculation,\n"
			"\tprint only split scopes and prepare output file\n"
		"  -t, --threads N\n"
			"\tsplit map calculation into N threads\n"
			"\tIf the number of columns or rows and threads is given, the requested number of threads is ignored.\n"
#endif
		"\nTIME, PIXELSIZE and AREA must be positive numbers; N must be a positive integer; WIDTH, HEIGHT, LEFT, and TOP must be integers.\n";



int main(int argc, char* argv[])
{
	int error;

	int xOffsetBinning;
	int yOffsetBinning;
	Scope* scopeList;
	CmdArgStruct params;

	// set error and help message parameters
	InitErrorHandling(ExtractAppName(argv[0]),helpText,true);

#ifdef DRAWAREA_MPI

	int NoP;
	unsigned long yOffset, height;
	unsigned long xOffset, width;
	unsigned long *heightList, *yOffsetList;
	unsigned long *widthList, *xOffsetList;

	// init MPI, get process id and total number of processes
	if (MPI_Init(&argc,&argv)!=MPI_SUCCESS)
		return Error(-1, "MPI_Init failed.\n");

	MPI_Comm_size(MPI_COMM_WORLD, &NoP);
	MPI_Comm_rank(MPI_COMM_WORLD, &ID);

	if (ID==MPI_ROOT_PROC)
	{

#endif // DRAWAREA_MPI

		// help requested?
		if (getCmdOption(argv+1,argv+argc,"-h") || getCmdOption(argv+1,argv+argc,"--help"))
			return Help();

		// evaluate cmd options
		if ((error=params.Evaluate(argc,argv))!=ERROR_SUCCESS)
			return error;

		// check if input file readable in PNM format
		PortableAnyMap inputMap(params.inputName,PNM_MODE_READ);
		if (!inputMap.IsOK())
		{
			return Error(inputMap.GetLastError(),"Input data file %s could not be read!\n%s\n",params.inputName,inputMap.GetLastErrorMessage());
		}


		params.scope.Fit(inputMap.GetSize());
		xOffsetBinning=((inputMap.GetWidth()-1)%params.nBinning)/2;
		yOffsetBinning=((inputMap.GetHeight()-1)%params.nBinning)/2;

		// check if output file is ready for updating (opens in update mode, has the write type and min size)
		// else try to create new file and open for update
		Scope outputScope((inputMap.GetWidth()-1)/params.nBinning+1,(inputMap.GetHeight()-1)/params.nBinning+1);
		PortableAnyMap outputMap(params.outputName,PNM_MODE_UPDATE);
		if (!outputMap.IsOK() || outputMap.GetType()!=PNM_TYPE_P5 || outputMap.GetWidth()<outputScope.Width() || outputMap.GetHeight()<outputScope.Height())
		{
			Error(ERROR_WARNING,"Output file %s not ready for update. Trying to recreate...\n",params.outputName);
			outputMap.Close();
			outputMap.Create(params.outputName,outputScope,PNM_TYPE_P5,65535);
			outputMap.Open(params.outputName,PNM_MODE_UPDATE);
			if (!outputMap.IsOK())
				return Error(outputMap.GetLastError(),"Failed to open output file %s.\n%s\n",params.outputName,outputMap.GetLastErrorMessage());
		}
		inputMap.Close();
		outputMap.Close();

#ifdef DRAWAREA_MPI
		params.nRows=NoP;
		params.nColumns=1;
		params.nThreads=NoP;
#endif // DRAWAREA_MPI

		// setup height and offset list for heuristic splitting
		scopeList=new Scope[params.nThreads];
		if (!scopeList)
			return Error(-1, MPI_MSG("Unable to allocate memory for heuristic splitting.\n"));

		// split map into params.nRows*params.nColumns scopes
		// heuristic will split in scopes with approx. equal calculation time
		if ((error=SplitHeuristic(params.inputName, params.scope, scopeList, params.nColumns, params.nRows, params.limitTime, params.pixelSize))!=ERROR_SUCCESS)
			return error;
		for (int i=0; i<params.nThreads; i++)
			Fit2Binning(scopeList[i],params.nBinning,xOffsetBinning,yOffsetBinning);

#ifdef DRAWAREA_MPI

		// prepare scopes for MPI scattering
		widthList=new unsigned long[params.nThreads];
		xOffsetList=new unsigned long[params.nThreads];
		heightList=new unsigned long[params.nThreads];
		yOffsetList=new unsigned long[params.nThreads];
		if (!widthList || !xOffsetList || !heightList || !yOffsetList)
			return Error(-1, MPI_MSG("Unable to allocate memory for mpi scattering of scopes.\n"));

		for (int i=0; i<params.nThreads; i++)
		{
			widthList[i]=scopeList[i].Width();
			xOffsetList[i]=scopeList[i].First(COLUMN);
			heightList[i]=scopeList[i].Height();
			yOffsetList[i]=scopeList[i].First(ROW);
		}
	}

	// send evaluated cmd args to all processes
	MPI_Bcast(&params.inputName, 256, MPI_UNSIGNED_CHAR, MPI_ROOT_PROC, MPI_COMM_WORLD);
	MPI_Bcast(&params.outputName, 256, MPI_UNSIGNED_CHAR, MPI_ROOT_PROC, MPI_COMM_WORLD);
	MPI_Bcast(&params.limitTime, 1, MPI_FLOAT, MPI_ROOT_PROC, MPI_COMM_WORLD);
	MPI_Bcast(&params.pixelSize, 1, MPI_FLOAT, MPI_ROOT_PROC, MPI_COMM_WORLD);
	MPI_Bcast(&params.areaScaling, 1, MPI_FLOAT, MPI_ROOT_PROC, MPI_COMM_WORLD);
	MPI_Bcast(&params.nBinning, 1, MPI_INT, MPI_ROOT_PROC, MPI_COMM_WORLD);
	MPI_Bcast(&xOffsetBinning, 1, MPI_INT, MPI_ROOT_PROC, MPI_COMM_WORLD);
	MPI_Bcast(&yOffsetBinning, 1, MPI_INT, MPI_ROOT_PROC, MPI_COMM_WORLD);

	// send calculated scopes to all processes
	MPI_Scatter(yOffsetList, 1, MPI_UNSIGNED_LONG, &yOffset, 1, MPI_UNSIGNED_LONG, MPI_ROOT_PROC, MPI_COMM_WORLD);
	MPI_Scatter(heightList, 1, MPI_UNSIGNED_LONG, &height, 1, MPI_UNSIGNED_LONG, MPI_ROOT_PROC, MPI_COMM_WORLD);
	MPI_Scatter(xOffsetList, 1, MPI_UNSIGNED_LONG, &xOffset, 1, MPI_UNSIGNED_LONG, MPI_ROOT_PROC, MPI_COMM_WORLD);
	MPI_Scatter(widthList, 1, MPI_UNSIGNED_LONG, &width, 1, MPI_UNSIGNED_LONG, MPI_ROOT_PROC, MPI_COMM_WORLD);

	// set process scope
	params.scope=Scope(width,height,xOffset,yOffset);

#else // ifdef DRAWAREA_MPI

	if (params.nThreads>1)
	{
		for (int i=0; i<params.nThreads; i++)
			printf("scope %d is %ux%u+%u+%u\n",i,scopeList[i].Width(),scopeList[i].Height(),scopeList[i].First(COLUMN),scopeList[i].First(ROW));
#ifdef DRAWAREA_THREADS
		if (params.noCalc)
#endif
			return 0;
	}

#endif // DRAWAREA_MPI

	// perform local calculation
	clock_t start=clock();
	CalculateAreaMap(params.inputName, params.outputName, params.scope, params.limitTime, params.pixelSize, params.areaScaling, params.nBinning);
	Info(MPI_MSG("map calculation took %.2f s\n"),float(clock()-start)/CLOCKS_PER_SEC);

#ifdef DRAWAREA_MPI
	MPI_Finalize();
#endif // DRAWAREA_MPI

	return 0;
}

// shrink scope to matches binning offsets
Scope& Fit2Binning(Scope& scope, int nBinning, int xOffset, int yOffset)
{
	size_t last;
	last=scope.Last(COLUMN)-(nBinning+scope.Last(COLUMN)%nBinning-xOffset)%nBinning;
	scope.First(COLUMN)+=(nBinning+xOffset-scope.First(COLUMN)%nBinning)%nBinning;
	scope.Width()=(last<scope.First(COLUMN)?0:last-scope.First(COLUMN)+1);

	last=scope.Last(ROW)-(nBinning+scope.Last(ROW)%nBinning-yOffset)%nBinning;
	scope.First(ROW)+=(nBinning+yOffset-scope.First(ROW)%nBinning)%nBinning;
	scope.Height()=(last<scope.First(ROW)?0:last-scope.First(ROW)+1);

	return scope;
}

// split a list of nValues values into nParts with approx. equal sum(values)
size_t* Split2EqualParts(size_t nValues, double* values, size_t nParts, size_t *parts)
{
	// get total weight
	double totalWeight=0.0;
	for (size_t i=0; i<nValues; i++)
		totalWeight+=values[i];

	// allocate memory for part split
	if (!parts)
		parts=new size_t[nParts+1];
	if (!parts)
		return NULL;

	size_t part=0;
	double partWeightActual=0.0;
	double partWeightSetPoint=totalWeight/(nParts-part);
	parts[part]=0;

	for (size_t i=0; i<nValues && part<nParts-1; i++)
	{
		// if not last remaining part, part contains at least 1 value,
		// and actual part weight plus half value exceeds set point
		// switch to new part
		if ((part<nParts-1) && (parts[part]<i) && (partWeightActual+values[i]/2>partWeightSetPoint))
		{
			parts[++part]=i;
			totalWeight-=partWeightActual;
			partWeightActual=0.0;
			partWeightSetPoint=totalWeight/(nParts-part);
		}
		// add value to part
		partWeightActual+=values[i];
	}

	// set remaining parts to end of value list
	while (part<nParts)
		parts[++part]=nValues;

	return parts;
}

// calculate the mean travel time for one pixel as mean of travel times to all valid neighbors
float MeanTravelTime(size_t *origin, size_t dataWidth, float pixelSize, int& validNeighbors)
{
	validNeighbors=0;

	// skip if pixel is invalid
	if (*origin==INVALID_ALTITUDE)
		return 0.0;
	
	float distance=0.0;
	float diagPixelSize=SQRT_2*pixelSize;
	size_t *neighbor;

	// sum up travel times to all valid neighbors
	for (int i=-1; i<2; i++)
	{
		neighbor=origin+i*dataWidth-1;
		for (int j=-1; j<2; j++)
		{
			if ( (i!=0 || j!=0) && *neighbor!=INVALID_ALTITUDE )
			{
				distance+=Edge::Distance(*origin,*neighbor,(i*j!=0?diagPixelSize:pixelSize));
				validNeighbors++;
			}
			neighbor++;
		}
	}

	// calculate mean value
	if (validNeighbors>0)
		distance/=validNeighbors;

	return distance;
}


// read input file and split into nColumns x nRows equally weighted scopes
// 	- the weight of each scope is the sum of the normalized approximated pixel times to the power of 4
// 	- the approximated pixel time is the mean time needed to get to a neighboring pixel
//	- normalized pixel time is MAX(1.0 - PIXEL_TIME / TIME_LIMIT, 0.0) times the number of valid neighbors
int SplitHeuristic(const char* inputName, Scope inputScope, Scope* scopeList, size_t nColumns, size_t nRows, float limitTime, float pixelSize)
{
	if (nColumns>1 || nRows>1)
		Info("heuristic map splitting ...\n");

	// check if scopeList is initialized
	if (!scopeList)
		return Error(-1, MPI_MSG("No memory allocated for scope list.\n"));

	// open input map
	PortableAnyMap inputMap(inputName,PNM_MODE_READ);
	if (!inputMap.IsOK())
		return Error(inputMap.GetLastError(),"Input data file %s could not be read!\n%s\n", inputName,inputMap.GetLastErrorMessage());

	inputScope.Fit(inputMap.GetSize());

	// allocate memory for map evaluation
	size_t *invalid=new size_t[inputScope.Width()+2];
	size_t *pixs=new size_t[3*inputScope.Width()+6];
	double *weights=new double[std::max(inputScope.Height(),inputScope.Width())];
	size_t *parts=new size_t[std::max(nRows,nColumns)+1];
	if (!invalid || !pixs || !weights || !parts)
		return Error(-1, MPI_MSG("Unable to allocate memory for heuristics.\n"));
	
	// initialize memory for map evaluation
	for (size_t i=0; i<inputScope.Width()+2; i++) 
		invalid[i]=INVALID_ALTITUDE;
	size_t *prevPixs=pixs+1;
	size_t *currPixs=pixs+inputScope.Width()+3;
	size_t *nextPixs=pixs+2*inputScope.Width()+5;
	memcpy(prevPixs-1,invalid,(inputScope.Width()+2)*sizeof(*pixs));
	memcpy(currPixs-1,invalid,(inputScope.Width()+2)*sizeof(*pixs));
	memcpy(nextPixs-1,invalid,(inputScope.Width()+2)*sizeof(*pixs));
	int validNeighbors;

	// row splitting first:
	if (nRows>1)
	{
		Info("\tanalyzing rows ");

		// get weight of each line in map and total map weight
		Scope readScope=inputScope;
		readScope.Height()=1;
		inputMap.ReadScope(readScope,nextPixs);
		for (size_t i=0; i<inputScope.Height(); i++)
		{
			weights[i]=0.0;
			readScope.First(ROW)++;
			memcpy(prevPixs,currPixs,inputScope.Width()*sizeof(*pixs));
			memcpy(currPixs,nextPixs,inputScope.Width()*sizeof(*pixs));

			size_t num=inputMap.ReadScope(readScope,nextPixs);
			if (num<inputScope.Width())
				memcpy(nextPixs+num,invalid,(inputScope.Width()-num)*sizeof(*pixs));

			// get weight of each pixel by calculating travel times to all neighbors
			for (size_t j=0; j<inputScope.Width(); j++)
			{
				float time=MeanTravelTime(currPixs+j,inputScope.Width(),pixelSize,validNeighbors);
				weights[i]+=PixelWeight(time,limitTime)*validNeighbors;
			}
			if (i%1000==0) Info(".");
		}
		Info(" done\n");
	}

	// split map into equally weighted rows
	Split2EqualParts(inputScope.Height(),weights,nRows,parts);
	for (size_t row=0; row<nRows; row++)
		for (size_t col=0; col<nColumns; col++)
			scopeList[row*nColumns+col].GetRange(ROW)=Range(parts[row+1]-parts[row],inputScope.First(ROW)+parts[row]);

	// reinitialize memory for map evaluation
	memcpy(prevPixs-1,invalid,(inputScope.Width()+2)*sizeof(*pixs));
	memcpy(currPixs-1,invalid,(inputScope.Width()+2)*sizeof(*pixs));
	memcpy(nextPixs-1,invalid,(inputScope.Width()+2)*sizeof(*pixs));

	// loop through all row bundles to split into columns
	for (size_t row=0; row<nRows; row++)
	{
		if (nColumns>1)
		{
			Info("\tanalyzing columns of row %d ",row+1);

			// get weight of each column in row bundle and total bundle weight
			Scope readScope=inputScope;
			readScope.Height()=1;
			readScope.First(ROW)=scopeList[row*nColumns].First(ROW);
			inputMap.ReadScope(readScope,nextPixs);
			for (size_t i=0; i<inputScope.Width(); i++)
				weights[i]=0.0;
			for (size_t i=0; i<scopeList[row*nColumns].Height(); i++)
			{
				readScope.First(ROW)++;
				memcpy(prevPixs,currPixs,inputScope.Width()*sizeof(*pixs));
				memcpy(currPixs,nextPixs,inputScope.Width()*sizeof(*pixs));

				size_t num=inputMap.ReadScope(readScope,nextPixs);
				if (num<inputScope.Width())
					memcpy(nextPixs+num,invalid,(inputScope.Width()-num)*sizeof(*pixs));

				// get weight of each pixel by calculating travel times to all neighbors
				for (size_t j=0; j<num; j++)
				{
					float time=MeanTravelTime(currPixs+j,inputScope.Width(),pixelSize,validNeighbors);
					weights[j]+=PixelWeight(time,limitTime)*validNeighbors;
				}
				if (i%1000==0) Info(".");
			}
			Info(" done\n");
		}

		// get splitting and save in scopes
		Split2EqualParts(inputScope.Width(),weights,nColumns,parts);
		for (size_t col=0; col<nColumns; col++)
			scopeList[row*nColumns+col].GetRange(COLUMN)=Range(parts[col+1]-parts[col],parts[col]+inputScope.First(COLUMN));
	}

	inputMap.Close();
	delete [] invalid;
	delete [] pixs;
	delete [] weights;
	delete [] parts;

	return ERROR_SUCCESS;
}

// read Node array scope from map file
// if scope is out of total map boundaries, adjust scope
Node* ReadScopeBinaryMatrix(const char* fileName, Scope& scope)
{
	// open input map
	PortableAnyMap input(fileName,PNM_MODE_READ);
	if (!input.IsOK())
	{
		Error(input.GetLastError(), MPI_MSG("Unable to open matrix file %s to read data.\n%s\n"), fileName, input.GetLastErrorMessage());
		return NULL;
	}

	// adjust scope to fit into map dimensions
	scope.Fit(input.GetSize());

	// allocate memory for Node array
	Node* nodes=new Node[scope.Size()];
	if (!nodes)
	{
		Error(-1, MPI_MSG("Unable to allocate memory for nodes.\n"));
		return NULL;
	}

	Info(MPI_MSG("reading nodes from scope %u x %u + %u + %u\n"),scope.Width(),scope.Height(),scope.First(COLUMN),scope.First(ROW));

	// read nodes scope from input map
	input.ReadScope(scope,nodes),scope.Size();

	// close input file
	input.Close();

	return nodes;
}

// calculate area in scope:
//	- create maximal reachable node map from
//	  input map file
//	- calculate area for each node in scope
//	  and write it to output map file
int CalculateAreaMap(const char* inputName, const char* outputName, Scope& scope, float limitTime, float pixelSize, float areaScaling, int nBinning)
{
	// open map file to update pixel data
	PortableAnyMap output(outputName,PNM_MODE_UPDATE);
	if (!output.IsOK())
		return Error(output.GetLastError(), MPI_MSG("Unable to open matrix file %s for data output.\n%s\n"), outputName, output.GetLastErrorMessage());

	// adjust scope, if necessary
	scope.Fit(output.GetSize()*nBinning);

	// enlarge scope by the maximal travel distance in TIME_LIMIT
	// to get read scope (all nodes that may be needed)
	Scope readScope=scope+int(limitTime/(pixelSize*MIN_TIME));

	// read nodes from binary file
	Node* nodes=ReadScopeBinaryMatrix(inputName,readScope);
	if (!nodes)
		return -1;

	// connect nodes
	CreateGraphEdges(nodes,readScope.Width(),readScope.Height(),pixelSize);

	// adjust scope if necessary
	scope.Fit(readScope);
	Scope writeScope((scope.Width()-1)/nBinning+1,1,scope.First(COLUMN)/nBinning,scope.First(ROW)/nBinning);

	// calculate offset between read nodes and scope nodes
	size_t nodesOffset=(scope.First(ROW)-readScope.First(ROW))*readScope.Width()
			+ scope.First(COLUMN)-readScope.First(COLUMN);

	// allocate memory to write one row at once
	size_t* area=new size_t[writeScope.Width()];
	if (!area)
		return Error(-1, MPI_MSG("No memory allocated for row data.\n"));

	Info(MPI_MSG("calculating area in scope %u x %u + %u + %u ...\n"),scope.Width(),scope.Height(),scope.First(COLUMN),scope.First(ROW));

	clock_t start;

	// loop all nodes in scope and
	// get area that could be reached within TIME_LIMIT
	for (size_t i=0;i<scope.Height();i+=nBinning)
	{
		// for each row in scope:
		// 	- calculate area for each node in row
		//	- write row to output file
		start=clock();
		for (size_t j=0;j<writeScope.Width();j++)
			area[j]=WayPoint(nodes+nodesOffset+i*readScope.Width()+j*nBinning).GetAreaDijkstra(limitTime)/areaScaling;

		output.UpdateScope(writeScope,area);
		writeScope.First(ROW)++;

		Info("\trow %u (%u/%u) done in %.2f s\n",i+scope.First(ROW),(i/nBinning)+1,(scope.Height()-1)/nBinning+1,float(clock()-start)/CLOCKS_PER_SEC);
	}

	output.Close();

	delete [] area;
	delete [] nodes;

	return ERROR_SUCCESS;
}
