/*********************************************************************

    Copyright 2016, Michael Steinert

    This file is part of DrawArea (Dijkstra Routing Algorithm Which
    Ascertains REachable Areas).

    DrawArea is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DrawArea is distributed WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DrawArea.  If not, see <http://www.gnu.org/licenses/>.

***********************************************************************/

#ifndef A_STAR_INCLUDED
#define A_STAR_INCLUDED

#include <vector>
#include <set>
#include <stdio.h>
#include <stdint.h>

#define INVALID_ALTITUDE 55537
#define MIN_TIME               1.8  	// s/m //0.72 old value
#define POSITIVE_SLOPE_FACTOR  15	// s/m/100%
#define NEGATIVE_SLOPE_FACTOR -5 	// s/m/-100%
#define SQRT_2 1.41421356237

// Node is the basic structure of the graph
struct Node {
	float altitude;					// the altitude at this position
	std::vector<struct Edge> neighbors;		// vector of edges to all nearest and next-nearest neighbors
	Node(size_t alt=0);
	void AddNeighbour(int shift, float distance);
	float MeanDistance();
};

// Edge is the basic structure to connect nodes of the graph
struct Edge {
	int shift;				// pointer shift to get from node to neighbor
	float distance;				// temporal distance between the neighboring nodes
	Edge(Node* start, int shift, float spacialDistance);
	static float Distance(float start, float finish, float spacialDistance);	// calculating temporal edge length
};

void CreateGraphEdges(Node* nodes, size_t numCols, size_t numRows, float distance);

// MapPoint is basically just a wrapper
// to list, sort and compare WayPoint pointers
struct MapPoint {
	class WayPoint* wayPoint;
	MapPoint(class WayPoint* wp);
};

bool operator<(const MapPoint& mp1, const MapPoint& mp2);

// WayPoint is a point on the map and stores the
// minimal time to get to the origin
class WayPoint {

    private:
	Node* node;
	float minOriginTime;		// currently known shortest time to the origin

    public:
	WayPoint(Node* node);
	~WayPoint();
	const WayPoint* GetRoute();
	void Expand(class OpenList* openList, class ClosedList* ClosedList, float limitTime);
	size_t GetAreaDijkstra(float limitTime);
	bool operator>(const WayPoint& wp);

	friend bool operator==(const MapPoint& mp, const WayPoint* wp);
};

// all nodes which are partially evaluated
// (a route to the origin is known,
// but not necessarily the shortest)
// are stored in an OpenList
class OpenList {

    private:
	std::vector<MapPoint> v;
	std::vector<MapPoint>::iterator it;

    public:
	OpenList();
	~OpenList();
	void Push(WayPoint* wp);
	WayPoint* Pop();
	void Update();
	size_t Size();
	bool Empty();
	bool Contains(WayPoint* wp);
	WayPoint* Select();

};

// all nodes which are fully evaluated
// (the shortest route to the origin is known)
// are stored in a ClosedList
class ClosedList {

    private:
	std::set<Node*> v;

    public:
	ClosedList();
	~ClosedList();
	void Push(Node* node);
	size_t Size();
	bool Empty();
	bool Contains(Node* node);
};

#endif	// A_STAR_INCLUDED
