/*********************************************************************

    Copyright 2016, Michael Steinert

    This file is part of DrawArea (Dijkstra Routing Algorithm Which
    Ascertains REachable Areas).

    DrawArea is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DrawArea is distributed WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DrawArea.  If not, see <http://www.gnu.org/licenses/>.

***********************************************************************/

#ifndef ERROR_H_INCLUDED
#define ERROR_H_INCLUDED

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#define ERROR_SUCCESS 0
#define ERROR_WARNING 0

#ifdef _WIN32
	#include <windows.h>
#endif

void InitErrorHandling(const char* appName, const char* heltText, bool consoleErrorOut);
int Error(const char* str, ...);
int Error(int errCode, const char* str, ...);
int Info(const char* str, ...);
int Help(const char* msg=NULL);
int Help(int code, const char* msg=NULL);

static const char* HELP_TEXT=NULL;
static const char* APP_NAME=NULL;
static bool CONSOLE_ERROR_OUT=true;

inline void InitErrorHandling(const char* appName, const char* helpText, bool consoleErrorOut)
{
	APP_NAME=appName;
	HELP_TEXT=helpText;
	CONSOLE_ERROR_OUT=consoleErrorOut;
}

// print error message
inline int ErrorVA(int code, const char* str, va_list pArg)
{
	va_list pArgCopy;
	va_copy(pArgCopy, pArg);
	// get length of error msg to be printed
	int len=vsnprintf(NULL,0,str,pArgCopy)+1;
	va_end(pArgCopy);

	// allocate memory and write error message to string
	char* msg=(char*)malloc(len);
	vsprintf(msg,str,pArg);

	// output error message
	if (CONSOLE_ERROR_OUT)
		fprintf(stderr,(code!=ERROR_SUCCESS?"Error: %s":"%s"),msg);
#ifdef _WIN32
	else MessageBox(NULL,msg,(code!=0?"Error":"Help"),MB_OK|(code!=0?MB_ICONERROR:MB_ICONINFORMATION));
#endif

	free(msg);
	return code;
}

// print error message wrapper
inline int Error(int code, const char* str, ...)
{
	va_list pArg;
	va_start(pArg,str);
	return ErrorVA(code,str,pArg);
}

// print error message wrapper with default error code 0
inline int Info(const char* str, ...)
{
	va_list pArg;
	va_start(pArg,str);
	return ErrorVA(ERROR_SUCCESS,str,pArg);
}

// print error message wrapper with default error code -1
inline int Error(const char* str, ...)
{
	va_list pArg;
	va_start(pArg,str);
	return ErrorVA(-1,str,pArg);
}

// create help message
inline int Help(int code, const char* msg)
{
	if (msg)
	{
		char* msgHelp=(char*)malloc(strlen(HELP_TEXT)+strlen(msg)+2);
		strcpy(msgHelp,msg);
		strcpy(msgHelp+strlen(msgHelp),"\n\n");
		strcpy(msgHelp+strlen(msgHelp),HELP_TEXT);
		code=Error(code,msgHelp,APP_NAME);
		free(msgHelp);
		return code;
	}
	return Error(code,HELP_TEXT,APP_NAME);
}

inline int Help(const char* msg)
{
	return Help(ERROR_SUCCESS,msg);
}

#endif // ERROR_H_INCLUDED
