/*********************************************************************

    Copyright 2016, Michael Steinert

    This file is part of DrawArea (Dijkstra Routing Algorithm Which
    Ascertains REachable Areas).

    DrawArea is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DrawArea is distributed WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DrawArea.  If not, see <http://www.gnu.org/licenses/>.

***********************************************************************/

#ifndef PARSECMDLARG_H
#define PARSECMDLARG_H

#include <stdlib.h>
#include <string.h>

/**********************************************************************

  This part is general purpose stuff to parse command line arguments.

***********************************************************************/

#ifdef _WIN32
	char PATH_SEPARATOR='\\';
#else
	char PATH_SEPARATOR='/';
#endif

bool getCmdOption(char** begin, char** end, const char* option);
bool getCmdOption(char** begin, char** end, const char* option, char* value);
bool getCmdOption(char** begin, char** end, const char* option, int& value);
bool getCmdOption(char** begin, char** end, const char* option, float& value);
char* ExtractAppName(const char* cmd);

inline char** find(char** begin, char** end, const char* target)
{
	char** itr=begin;
	while (itr<end && strcmp(*itr,target)!=0 )  ++itr;
	return itr;
}

inline bool getCmdOption(char** begin, char** end, const char* option)
{
	return find(begin,end,option)!=end;
}

inline bool getCmdOption(char** begin, char** end, const char* option, char* value)
{
	char** itr=find(begin,end,option);
	if ( (itr==end) || (++itr==end) ) return false;
	strcpy(value,*itr);
	return true;
}

inline bool getCmdOption(char** begin, char** end, const char* option, int& value)
{
	char str[1024];
	char* error;
	if (!getCmdOption(begin,end,option,str)) return false;
	value=strtol(str,&error,10);
	return *error==0;
}

inline bool getCmdOption(char** begin, char** end, const char* option, float& value)
{
	char str[1024];
	char* error;
	if (!getCmdOption(begin,end,option,str)) return false;
	value=strtof(str,&error);
	return *error==0;
}

inline char* ExtractAppName(const char* cmd)
{
	const char *fileName=cmd;
	char *path=strrchr(fileName,PATH_SEPARATOR);		// get start pointer of filename w/o path
	if (path) fileName=path+1;				// set filename start
	int cCount=strlen(fileName);				// get length of file name

	char* ext=strrchr(fileName,'.');			// get start pointer of old extension
	if (ext && ext!=fileName) cCount=ext-fileName;		// get length of file name w/o extension

	char* str=new char[cCount+1];				// create char array to store file name and
	strncpy(str,fileName,cCount);				// copy old file name w/o extension
	str[cCount]='\0';

	return str;
}

/**********************************************************************

  This part is specially designed for DrawArea.

***********************************************************************/

#include "Error.h"
#include "pnm.h"

struct CmdArgStruct
{
	Scope scope;
	float areaScaling;
	int nBinning;
	int nColumns;
	float limitTime;
	float pixelSize;
	bool noCalc;
	int nRows;
	int nThreads;
	char inputName[256];
	char outputName[256];

	int Evaluate(int argc, char** argv);
};

// evaluated command line arguments
inline int CmdArgStruct::Evaluate(int argc, char** argv)
{

/* mandatory arguments*/

	// set input file
	if (argc<2)
		return Help(-1,"Missing input file.");
	strcpy(inputName,argv[argc-1]);

	// set time limit
	if (getCmdOption(argv+1,argv+argc,"-l",limitTime) || getCmdOption(argv+1,argv+argc,"--limit",limitTime))
	{
		if (!limitTime>0)
			return Help(-1,"Travel time limit must be a positive.");
	}
	else return Help(-1,"Missing travel time limit.");

	// set pixel size
	if (getCmdOption(argv+1,argv+argc,"-p",pixelSize) || getCmdOption(argv+1,argv+argc,"--pixel-size",pixelSize))
	{
		if (!pixelSize>0)
			return Help(-1,"Pixel size must be a positive.");
	}
	else return Help(-1,"Missing image pixel size.");

/* optional arguments*/

	// set area scaling
	areaScaling=1.0;
	if (getCmdOption(argv+1,argv+argc,"-a",areaScaling) || getCmdOption(argv+1,argv+argc,"--area",areaScaling))
		if (!areaScaling>0)
			return Help(-1,"Area scaling factor must be a positive.");
	areaScaling*=1E6/pixelSize/pixelSize;

	// set binning
	nBinning=1;
	if (getCmdOption(argv+1,argv+argc,"-b",nBinning) || getCmdOption(argv+1,argv+argc,"--binning",nBinning))
		if (nBinning<1)
			return Help(-1,"Binning must be a positive.");
	areaScaling*=nBinning*nBinning;

	// set number of columns to split into
	nColumns=0;
	if (getCmdOption(argv+1,argv+argc,"-c") || getCmdOption(argv+1,argv+argc,"--cols"))
	{
		nColumns=-1;
		if (getCmdOption(argv+1,argv+argc,"-c",nColumns) || getCmdOption(argv+1,argv+argc,"--cols",nColumns))
			if (nColumns<1)
				return Help(-1,"Number of columns must be a positive.");
	}

	// user requested only scope splitting but no map calculation
	noCalc=false;
	if (getCmdOption(argv+1,argv+argc,"-n") || getCmdOption(argv+1,argv+argc,"--no-calc"))
		noCalc=true;

	// set output name
	if (!getCmdOption(argv+1,argv+argc,"-o",outputName) && !getCmdOption(argv+1,argv+argc,"--output",outputName))
	{
		strcpy(outputName,"area_of_\0");
		strcpy(outputName+strlen(outputName),inputName);
	}

	// set number of rows to split into
	nRows=0;
	if (getCmdOption(argv+1,argv+argc,"-r") || getCmdOption(argv+1,argv+argc,"--rows"))
	{
		nRows=-1;
		if (getCmdOption(argv+1,argv+argc,"-r",nRows) || getCmdOption(argv+1,argv+argc,"--rows",nRows))
			if (nRows<1)
				return Help(-1,"Number of rows must be a positive.");
	}

	// set user defined scope
	scope=Scope(-1,-1);
	char strScope[256];
	if (getCmdOption(argv+1,argv+argc,"-s",strScope) || getCmdOption(argv+1,argv+argc,"--scope",strScope))
		if (sscanf(strScope,"%ldx%ld+%ld+%ld",&scope.Width(),&scope.Height(),&scope.First(COLUMN),&scope.First(ROW))!=4)
			return Help(-1,"Scope must be of format INTxINT+INT+INT.");

	// set number of threads
	nThreads=1;
	if (getCmdOption(argv+1,argv+argc,"-t",nThreads) || getCmdOption(argv+1,argv+argc,"--threads",nThreads))
		if (nThreads<1)
			return Help(-1,"Number of threads must be a positive.");

	// calculate number of rows and columns to split into
	if (nRows>0 && nColumns>0)
		nThreads=nRows*nColumns;
	else
	{
		if (nColumns)
		{
			nRows=1;
			nThreads=nColumns=(nColumns>0?nColumns:nThreads);
		}
		else
		{
			nColumns=1;
			nThreads=nRows=(nRows>0?nRows:nThreads);
		}
	}

	printf("cols: %d\nrows: %d\n",nColumns,nRows);

	return ERROR_SUCCESS;
}

#endif
