/*********************************************************************

    Copyright 2016, Michael Steinert

    This file is part of DrawArea (Dijkstra Routing Algorithm Which
    Ascertains REachable Areas).

    DrawArea is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DrawArea is distributed WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DrawArea.  If not, see <http://www.gnu.org/licenses/>.

***********************************************************************/

#include "DrawArea.h"
#include <algorithm>

// constructor:
// 	connect the MapPoint with a WayPoint
MapPoint::MapPoint(WayPoint* wp)
{
	wayPoint=wp;
}

// compare two MapPoints
// this is needed to heap the open list
bool operator<(const MapPoint& mp1, const MapPoint& mp2)
{
    return *(mp1.wayPoint)>*(mp2.wayPoint);
}

// constructor:
// 	connect the WayPoint with a node
//
// this is quite useful because with this
// you can use a Node pointer in functions
// that require a WayPoint
WayPoint::WayPoint(Node* node)
{
	this->node=node;
	minOriginTime=0.0;
}

// destructor (see constructor ClosedList)
WayPoint::~WayPoint() {}

// evaluate all neighboring nodes by estimating time to travel through this nodes
void WayPoint::Expand(OpenList* openList, ClosedList* closedList, float limitTime)
{
	// loop through all neighboring nodes
	for (std::vector<Edge>::iterator it=node->neighbors.begin(); it!=node->neighbors.end(); ++it)
	{
		// if current neighbor is fully evaluated (contained in closedList)
		// continue with the next neighbor
		if (closedList->Contains(node+it->shift))
			continue;

		WayPoint currNeighbour(node+it->shift);
		WayPoint* pCurrNeighbour=&currNeighbour;

		// check if node is already contained in openList
		// if so, use WayPoint in openList
		bool inOpenList=openList->Contains(pCurrNeighbour);
		if (inOpenList)
			pCurrNeighbour=openList->Select();

		// calculate time to origin as sum of previous node time and
		// time to get from there to this node
		float originTime=minOriginTime+it->distance;

		// if time limit is exceeded continue with next neighbor
		if (originTime>limitTime)
			continue;

		// if shorter route is known continue with next neighbor
		if (inOpenList && originTime>=pCurrNeighbour->minOriginTime)
			continue;

		// refresh times and shortest route previous node
		pCurrNeighbour->minOriginTime=originTime;

		// update openList
		if (inOpenList)
			openList->Update();
		else
			openList->Push(pCurrNeighbour);
	}
}

// this function is looking for all nodes which could be reached in time limit
// by using the Dijkstra algorithm
size_t WayPoint::GetAreaDijkstra(float limitTime)
{
	if (node->altitude==INVALID_ALTITUDE)
		return 0;

	minOriginTime=0.0;

	// define empty closed and open list
	ClosedList closedList;
	OpenList openList;

	// start at origin
	openList.Push(this);

	// loop until all nodes that could be reached in time limit are fully evaluated
	while (!openList.Empty())
	{
		// use the WayPoint with the lowest travel time
		WayPoint* top=openList.Pop();

		// add the WayPoint to the closed list to mark it as fully evaluated
		closedList.Push(top->node);

		// evaluate all neighboring nodes
		top->Expand(&openList,&closedList,limitTime);

		delete top;
	}

	return closedList.Size();
}

bool WayPoint::operator>(const WayPoint& wp)
{
    return minOriginTime>wp.minOriginTime;
}

// check if MapPoint node and a Node pointer pointing to the same Node
bool operator==(const MapPoint& mp, const WayPoint* wp)
{
	return wp->node==mp.wayPoint->node;
}
