/*********************************************************************

    Copyright 2016, Michael Steinert

    This file is part of DrawArea (Dijkstra Routing Algorithm Which
    Ascertains REachable Areas).

    DrawArea is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DrawArea is distributed WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DrawArea.  If not, see <http://www.gnu.org/licenses/>.

***********************************************************************/

#ifndef PNM_H_INCLUDED
#define PNM_H_INCLUDED

#include <stdio.h>
#include <string.h>
#include <algorithm>

// reading, writing, and (if binary P4-P6) updating
// portable anymap images

namespace {
// Range describes a numbered interval
class Range {

    private:
	size_t start;
	size_t length;

    public:
	Range(size_t length=0, size_t start=0) { this->length=length; this->start=start; }
	size_t& Size() { return length; }
	size_t& First() { return start; }
	size_t Last() { return start+length-1; }

	Range& Fit(const Range limits)
		{
			size_t last=std::min(start+length-1,limits.start+limits.length-1);
			start=std::max(start,limits.start);
			length=(last<start?(size_t)0:last-start+1);
			return *this;
		}
	bool IsInside(size_t value) { return (value>=start && value<start+length); }

	Range& operator<<=(int n) { if (n<0) return *this>>=n; start=(start>(size_t)n?start-n:0); return *this; }
	Range& operator>>=(int n) { if (n<0) return *this<<=n; start+=n; return *this; }
	Range& operator+=(int n)
		{
			if (n<0) return *this-=n;
			size_t oldLast=this->Last();
			*this<<=n;
			length=oldLast-start+n+1;
			return *this;
		}
	Range& operator-=(int n)
		{
			if (n<0) return *this+=n;
			if (length-2*n<0)
			{
				start=start+length/2;
				length=0;
			}
			else
			{
				*this>>=n;
				length-=2*n;
			}
			return *this;
		}
	Range& operator*=(double x) { if (x<0) x=-x; length*=x; return *this; }
	Range& operator/=(double x) { if (x<0) x=-x; length/=x; return *this; }

	Range operator<<(int n) { Range r=*this; return r<<=n; }
	Range operator>>(int n) { Range r=*this; return r>>=n; }
	Range operator+(int n) { Range r=*this; return r+=n; }
	Range operator-(int n) { Range r=*this; return r-=n; }
	Range operator*(double x) { Range r=*this; return r*=x; }
	Range operator/(double x) { Range r=*this; return r/=x; }

	bool operator==(Range r) { return start==r.start && length==r.length; }
	bool operator!=(Range r) { return start!=r.start || length!=r.length; }
};

// Scope is a 2D Range
enum DIMENSION {COLUMN = 0, ROW = 1};
class Scope {

    private:
	Range x,y;

    public:
	Scope(Range x, Range y) { this->x=x; this->y=y; }
	Scope(size_t xLength=0, size_t yLength=0, size_t xStart=0, size_t yStart=0) { x=Range(xLength,xStart); y=Range(yLength,yStart); }

	Range& GetRange(DIMENSION d) { if (d==ROW) return y; return x; }
	size_t Size() { return x.Size()*y.Size(); }
	size_t& Height() { return y.Size(); }
	size_t& Width() { return x.Size(); }
	size_t& First(DIMENSION d) { if (d==ROW) return y.First(); return x.First(); }
	size_t Last(DIMENSION d) { if (d==ROW) return y.Last(); return x.Last(); }

	Scope& Fit(const Scope limits) { x.Fit(limits.x); y.Fit(limits.y); return *this; }
	bool IsInside(size_t x, size_t y) { return this->x.IsInside(x) && this->y.IsInside(y); }

	Scope& operator+=(int n) { this->x+=n; this->y+=n; return *this; }
	Scope& operator-=(int n) { this->x-=n; this->y-=n; return *this; }
	Scope& operator*=(double x) { this->x*=x; this->y*=x; return *this; }
	Scope& operator/=(double x) { this->x/=x; this->y/=x; return *this; }

	Scope operator+(int n) { Scope s=*this; return s+=n; }
	Scope operator-(int n) { Scope s=*this; return s-=n; }
	Scope operator*(double x) { Scope s=*this; return s*=x; }
	Scope operator/(double x) { Scope s=*this; return s/=x; }

	bool operator==(Scope s) { return x==s.x && y==s.y; }
	bool operator!=(Scope s) { return x!=s.x || y!=s.y; }
};
}

namespace {

//  Magic Number |  Type   | Coding
//-----------------------------------
//       P1      | Bitmap  | ACSII
//       P2      | Graymap | ACSII
//       P3      | Pixmap  | ACSII
//       P4      | Bitmap  | binary
//       P5      | Graymap | binary
//       P6      | Pixmap  | binary
//
enum PNM_TYPE { PNM_TYPE_P1=1, PNM_TYPE_P2=2, PNM_TYPE_P3=3, PNM_TYPE_P4=4, PNM_TYPE_P5=5, PNM_TYPE_P6=6 };
enum PNM_MODE { PNM_MODE_NONE=0, PNM_MODE_READ, PNM_MODE_WRITE, PNM_MODE_UPDATE };

class PortableAnyMap {

    private:
    	Scope scope;
    	size_t maxValue;
    	size_t pixelSize;
	size_t rowSize;
    	PNM_TYPE type;
    	PNM_MODE mode;
    	size_t headerSize;
    	FILE* file;

    	size_t lineCount;
    	int lastError;
    	static const char* const pnmErrorList[];

    private:
    	void Clear();
    	bool SetError(int error);
	bool CheckBeforeReading(void* data);
	bool CheckBeforeUpdating(void* data);
	bool CheckBeforeWriting(void* data);
	bool DefineSizes();
	bool SkipComments();
	bool SkipASCIIValues(size_t n);

    public:
	PortableAnyMap();
	PortableAnyMap(const char* fileName, PNM_MODE mode);
	PortableAnyMap(const char* fileName, Scope scope, PNM_TYPE type, size_t maxValue=65535);
	~PortableAnyMap();

	bool Open(const char* fileName, PNM_MODE mode);
	bool Create(const char* fileName, size_t width, size_t height, PNM_TYPE type, size_t maxValue=65535);
	bool Create(const char* fileName, Scope scope, PNM_TYPE type, size_t maxValue=65535);
	bool Close();
	bool IsOK();
	int GetLastError(const char** errorMessage=NULL);
	const char* GetLastErrorMessage();

	Scope GetSize();
	size_t GetHeight();
	size_t GetWidth();
	size_t GetMaxValue();
	PNM_TYPE GetType();
    	PNM_MODE GetMode();
	void ShowInfo();

	template <typename T>
	size_t ReadScope(Scope& readScope, T* data, T (*explicitCast)(size_t)=NULL);
	template <typename T>
	size_t ReadLine(size_t line, T* data, T (*explicitCast)(size_t)=NULL);

	template <typename T>
	size_t UpdateScope(Scope& updateScope, T* data, size_t (*explicitCast)(T)=NULL);
	template <typename T>
	size_t UpdateLine(size_t line, T* data, size_t (*explicitCast)(T)=NULL);

	template <typename T>
	size_t WriteLine(T* data, size_t (*explicitCast)(T)=NULL);
	bool FillUp(size_t data_r=0, size_t data_g=-1, size_t data_b=-1);
};


/*************************************************************************************

	And now ... the definition part (because of the templates not in cpp file)

**************************************************************************************/

const char* const PortableAnyMap::pnmErrorList[] =
{
	"Everything is OK.",					// 0
	"Constructor doesn't match match file access mode.",	// 1
	"Failed on closing file.",				// 2
	"Operation permitted by file access mode.",		// 3
	"No memory allocated for data output.",			// 4
	"Invalid image dimensions.",				// 5
	"Invalid map type.",					// 6
	"Failed on opening file.",				// 7
	"Unexpectedly sudden end of file.",			// 8
	"Invalid file header.",					// 9
	"Update mode is not supported for ASCII coded maps.",	//10
	"Invalid data value.",					//11
	"An unknown error has occurred! PANIC!"
};

// empty constructor zeros all member variables
inline PortableAnyMap::PortableAnyMap()
{
	mode=PNM_MODE_NONE;
    	file=NULL;
    	lastError=0;
}

// read and update constructor opens file
inline PortableAnyMap::PortableAnyMap(const char* fileName, PNM_MODE mode)
{
	this->mode=PNM_MODE_NONE;
    	file=NULL;
	Open(fileName,mode);
}

// write constructor will create file
inline PortableAnyMap::PortableAnyMap(const char* fileName, Scope scope, PNM_TYPE type, size_t maxValue)
{
	mode=PNM_MODE_NONE;
    	file=NULL;
	Create(fileName,scope,type,maxValue);
}

// destruct by closing file if file pointer is still valid
inline PortableAnyMap::~PortableAnyMap()
{
	if (file) Close();
}

// query member variable values
inline Scope PortableAnyMap::GetSize() { return scope; }
inline size_t PortableAnyMap::GetHeight() { return scope.Height(); }
inline size_t PortableAnyMap::GetWidth() { return scope.Width(); }
inline size_t PortableAnyMap::GetMaxValue() { return maxValue; }
inline PNM_TYPE PortableAnyMap::GetType() { return type; }
inline PNM_MODE PortableAnyMap::GetMode() { return mode; }

// print map infos
inline void PortableAnyMap::ShowInfo()
{
	if (mode==PNM_MODE_NONE)
		printf("Portable anymap has not been opened yet. Mode is %s\n",(mode==PNM_MODE_NONE?"none":"unknown"));
	else
		printf("P%d\n%u %u\n%u\nopened in %s mode\n",
			type,
			scope.Width(), scope.Height(),
			maxValue,
			(mode==PNM_MODE_READ?"read":(mode==PNM_MODE_WRITE?"write":(mode==PNM_MODE_UPDATE?"update":"some unknown"))));
}


// set recent error
inline bool PortableAnyMap::SetError(int error)
{
	lastError=error;
	// define error that are not critical
	// on critical errors close fie and reset file pointer
	switch (error)
	{
		case 0: case 3: case 4:
			break;

		default:
			if (file)
				fclose(file);
			file=NULL;
			mode=PNM_MODE_NONE;
	}
	return (error==0);
}

// return last error and if argument pointer is valid
// point to error message
inline int PortableAnyMap::GetLastError(const char** errorMessage)
{
	if (errorMessage)
		*errorMessage=pnmErrorList[lastError];
	return lastError;
}
// return pointer to last error message
inline const char* PortableAnyMap::GetLastErrorMessage()
{
	return pnmErrorList[lastError];
}

// define rowSize and pixelSize depending on member variables
inline bool PortableAnyMap::DefineSizes()
{
	// in case of ASCII type define:
	// 	- rowSize as number of pixels in one row
	// 	- pixelSize as number of channels
	// in case of binary type define:
	//	- rowSize as size of one row in bytes
	//	- pixelSize as size of one pixel in bytes
	//	  or 0 in case of bitmap
	switch (type)
	{
		case PNM_TYPE_P1:
		case PNM_TYPE_P2:
			pixelSize=1;
			break;

		case PNM_TYPE_P3:
			pixelSize=3;
			break;

		case PNM_TYPE_P4:
			pixelSize=0;
			break;

		case PNM_TYPE_P5:
		case PNM_TYPE_P6:
			pixelSize=(maxValue>>8?2:1)*(type==PNM_TYPE_P5?1:3);
			break;

		default:;
	}

	if (type==PNM_TYPE_P4)
		rowSize=(scope.Width()+7)/8;
	else
		rowSize=scope.Width()*pixelSize;

	return true;
}

//skip comments that may be in the file header
inline bool PortableAnyMap::SkipComments()
{
	char c=getc(file);

	// skip white space
	while (feof(file) && (c==' ' || c=='\t' || c=='\r' || c=='n'))
		c=getc(file);
	if (feof(file))	return false;

	// skip comments
	while (getc(file)=='#')
	{
		// move file position to next line
		do c=getc(file);
		while (c!='\n' && c!='\r' && !feof(file));
		if (feof(file))
			return false;
	}

	fseek(file,-1,SEEK_CUR);
	return true;
}

// open file in read or update mode
// read header to set member variables
inline bool PortableAnyMap::Open(const char* fileName, PNM_MODE mode)
{
	if (file)
		Close();

	if (mode!=PNM_MODE_READ && mode!=PNM_MODE_UPDATE)
		return SetError(3);
	this->mode=mode;

	// set file access mode string for fopen()
	char strMode[]="rb+";
	if (mode==PNM_MODE_READ)
		strMode[2]='\0';

	// open file for output
	file=fopen(fileName,strMode);
	if (!file)
		return SetError(7);

	// read file header:
	//	- Magic Number
	//	- width
	//	- height
	//	- if not bitmap maxValue
	// all of the values are separated by whitespace and possibly comments
	// which is caught by SkipComments()
	if (fscanf(file,"P%d",&type)!=1 || type<PNM_TYPE_P1 || type>PNM_TYPE_P6)
		return SetError(6);
	if (!SkipComments() || fscanf(file,"%u",&scope.Width())!=1)
		return SetError(9);
	if (!SkipComments() || fscanf(file,"%u",&scope.Height())!=1)
		return SetError(9);
	if (type%3!=1)
	{
		if (!SkipComments() || fscanf(file,"%u",&maxValue)!=1)
			return SetError(9);
	}
	else
		maxValue=1;

	// if in update mode, map coding must be binary
	if (mode==PNM_MODE_UPDATE && type<PNM_TYPE_P4)
		return SetError(10);

	headerSize=ftell(file)+1;

	// check for valid map dimensions
	if (scope.Size()==0)
		return SetError(5);

	lastError=0;
	return DefineSizes();
}

// create file for write mode output
// write header and set member variables
inline bool PortableAnyMap::Create(const char* fileName, Scope scope, PNM_TYPE type, size_t maxValue)
{
	return Create(fileName,scope.Width(),scope.Height(),type,maxValue);
}

// create file for write mode output
// write header and set member variables
inline bool PortableAnyMap::Create(const char* fileName, size_t width, size_t height, PNM_TYPE type, size_t maxValue)
{
	// check if anymap is already assigned to a file
	if (file)
		Close();

	// check parameters for validity
	if (width==0 || height==0)
		return SetError(5);
	if (type<PNM_TYPE_P1 || type>PNM_TYPE_P6)
		return SetError(6);
	this->maxValue=std::min(size_t(65535),maxValue);
	if (type%3==1)
		this->maxValue=1;

	// open file for output
	file=fopen(fileName,"wb");
	if (!file)
		return SetError(7);

	// write header:
	// 	Pn  (magic number)
	// 	width height
	// 	maxValue (if not bitmap)
	fprintf(file,"P%d\n%u %u\n",type,width,height);
	if (type%3!=1)
		fprintf(file,"%u\n",maxValue);

	// set member variables
	scope=Scope(width,height);
	this->type=type;
	this->mode=PNM_MODE_WRITE;

	// reset lineCount to count written lines
	// to be able to fill image on close()
	lineCount=0;
	lastError=0;

	return 	DefineSizes();
}

// if file pointer is still valid
// close map by fill up pixels if in write mode
// and close file and reset file pointer
inline bool PortableAnyMap::Close()
{
	bool success=true;
	if (file)
	{
		if (mode==PNM_MODE_WRITE)
			success=FillUp();

		if (fclose(file)!=0)
		{
			success=false;
			lastError=2;
		}
	}

	file=NULL;
	mode=PNM_MODE_NONE;
	return success;
}

// if no error occurred, PNM is ok
inline bool PortableAnyMap::IsOK()
{
	return (lastError==0);
}

// check if file access mode is valid for reading (PNM_MODE_READ || PNM_MODE_UPDATE)
// and data pointer is set
inline bool PortableAnyMap::CheckBeforeReading(void* data)
{
	if (mode!=PNM_MODE_READ && mode!=PNM_MODE_UPDATE)
		return SetError(3);

	if (!data)
		return SetError(4);

	return true;
}

// check if file access mode is valid for updating (PNM_MODE_UPDATE)
// and data pointer is set
inline bool PortableAnyMap::CheckBeforeUpdating(void* data)
{
	if (mode!=PNM_MODE_UPDATE)
		return SetError(3);

	if (!data)
		return SetError(4);

	return true;
}

// check if file access mode is valid writing (PNM_MODE_WRITE)
// and data pointer is set
inline bool PortableAnyMap::CheckBeforeWriting(void* data)
{
	if (mode!=PNM_MODE_WRITE)
		return SetError(3);

	if (!data)
		return SetError(4);

	return true;
}

// skip n pixels in ascii coding
inline bool PortableAnyMap::SkipASCIIValues(size_t n)
{
	for (size_t i=0; i<n && !feof(file); i++)
		fscanf(file,"%*d");
	if (feof(file))
		return SetError(8);
	return true;
}

// read scope from map and write to preallocated data array
// and return number of data being read
template <typename T>
size_t PortableAnyMap::ReadScope(Scope& readScope, T* data, T (*explicitCast)(size_t))
{
	if (!CheckBeforeReading(data))
		return 0;

	if (!scope.IsInside(readScope.First(COLUMN),readScope.First(ROW)))
		return 0;

	readScope.Fit(scope);
	size_t dataCount=0;
	size_t val;

	// in case of ascii coding
	if (type<4)
	{
		// number of data values before and after scope in line
		size_t lineSuffix=scope.Last(COLUMN)-readScope.Last(COLUMN);
		size_t lineScope=readScope.Width()*pixelSize;
		size_t linePrefix=readScope.First(COLUMN)*pixelSize;

		// set position in file to first data value
		fseek(file,headerSize,SEEK_SET);

		// skip all lines before scope
		if (!SkipASCIIValues(rowSize*readScope.First(ROW)))
			return 0;

		// read line by line
		for (size_t i=0; i<readScope.Height(); i++)
		{
			// skip preceding columns
			if (!SkipASCIIValues(linePrefix))
				return 0;

			// read line from reading scope value by value and
			// write it to data array
			for (size_t j=0; j<lineScope && !feof(file); j++)
			{
				if (fscanf(file,"%u",&val)!=1)
				{
					SetError(11);
					return 0;
				}
				data[dataCount++]=(explicitCast?explicitCast(val):(T)val);
			}

			// skip succeeding columns
			if (!SkipASCIIValues(lineSuffix))
				return 0;
		}
	}
	else // in case of binary coding
	{
		int bitOffset=0;
		int linePrefix=readScope.First(COLUMN)*pixelSize;
		int lineScope=readScope.Width()*pixelSize;

		// in case of bitmap
		if (pixelSize==0)
		{
			lineScope=((readScope.Last(COLUMN)+7)/8-readScope.First(COLUMN)/8);
			bitOffset=readScope.First(COLUMN)%8;
			linePrefix=readScope.First(COLUMN)/8;
		}

		// position of first pertinent entry
		size_t offset=headerSize+rowSize*readScope.First(ROW)+linePrefix;

		// allocate memory tho read scope row at once
		unsigned char* row=new unsigned char[lineScope];

		// read line by line
		for (size_t i=0; i<readScope.Height(); i++)
		{
			// go to first pertinent line entry
			fseek(file,offset+i*rowSize,SEEK_SET);
			fread(row,1,lineScope,file);

			if (feof(file))
			{
				SetError(8);
				delete [] row;
				return 0;
			}

			// evaluate line value by value and
			// write it to data array
			unsigned char* rowPos=row;
			for (size_t j=0;j<readScope.Width();j++)
			{
				// for pixmap loop through colors
				for (size_t k=0; k<(pixelSize<3?1:3); k++)
				{
					if (pixelSize==0) // bitmap
					{
						int bitPos=(7-((bitOffset+j)&7));
						val=(*rowPos>>bitPos)&1;
						if (bitPos==0)
							rowPos++;
					}
					else	// pixmap or graymap
					{
						val=(pixelSize%2==1?0:*(rowPos++)<<8);
						val+=*(rowPos++);
					}
					data[dataCount++]=(explicitCast?explicitCast(val):(T)val);
				}
			}
		}
		delete [] row;
	}

	return dataCount;
}

// read line from map and write to preallocated data array
// and return number of data being read
template <typename T>
size_t PortableAnyMap::ReadLine(size_t line, T* data, T (*explicitCast)(size_t))
{
	Scope readScope(scope.Width(),1,0,line);
	return ReadScope(readScope, data, explicitCast);
}

// update scope with data array and return number of data being updated
template <typename T>
size_t PortableAnyMap::UpdateScope(Scope& updateScope, T* data, size_t (*explicitCast)(T))
{
	if (!CheckBeforeUpdating(data))
		return 0;

	if (!scope.IsInside(updateScope.First(COLUMN),updateScope.First(ROW)))
		return 0;

	Scope writeScope=updateScope;
	writeScope.Fit(scope);

	size_t dataCount=0;
	size_t val;
	T* pData=data;

	size_t bitOffset=0;
	unsigned char probeLast=0, probeFirst=0;
	size_t linePrefix=writeScope.First(COLUMN)*pixelSize;
	size_t lineScope=writeScope.Width()*pixelSize;

	// in case of bitmap
	if (pixelSize==0)
	{
		lineScope=(writeScope.Last(COLUMN)/8-writeScope.First(COLUMN)/8)+1;
		bitOffset=writeScope.First(COLUMN)%8;
		linePrefix=writeScope.First(COLUMN)/8;
		probeFirst=255-(255>>(writeScope.First(COLUMN)%8));
		probeLast=255>>(writeScope.Last(COLUMN)%8+1);
		if (lineScope<2)
			probeFirst|=probeLast;
	}

	// position of first pertinent entry
	size_t offset=headerSize+rowSize*writeScope.First(ROW)+linePrefix;

	// allocate memory tho read scope row at once
	unsigned char* row=new unsigned char[lineScope];

	// read line by line
	for (size_t i=0; i<writeScope.Height(); i++)
	{
		// in case of bitmap make sure, you don't change pixels
		// outside the scope at the edges
		if (pixelSize==0)
		{
			fseek(file,offset+i*rowSize,SEEK_SET);
			fread(row,1,lineScope,file);
			if (lineScope>2)
				memset(row+1,0,lineScope-2);
			*row&=probeFirst;
			if (lineScope>1)
				*(row+lineScope-1)&=probeLast;

		}

		pData=data+i*updateScope.Width();

		// evaluate line value by value and
		// write it to data array
		unsigned char* rowPos=row;
		for (size_t j=0;j<writeScope.Width();j++)
		{
			// for pixmap loop through colors
			for (size_t k=0; k<(pixelSize<3?1:3); k++)
			{

				val=std::min((size_t)65535,(explicitCast?explicitCast(*pData):(size_t)*pData));
				if (pixelSize==0) // bitmap
				{
					int bitPos=7-(bitOffset+j)%8;
					if (val)
						*rowPos|=1<<bitPos;
					if (bitPos==0)
						rowPos++;
				}
				else	// pixmap or graymap
				{
					if (pixelSize%2==0)
					{
						*(rowPos++)=(unsigned char)(val>>8);
						*(rowPos++)=(unsigned char)(val);
					}
					else
						*(rowPos++)=std::min((size_t)255,val);
				}
				pData++;
			}
		}

		// go to first pertinent line entry and write line
		fseek(file,offset+i*rowSize,SEEK_SET);
		fwrite(row,1,lineScope,file);
		if (feof(file))
		{
			SetError(8);
			delete [] row;
			return 0;
		}
	}

	delete [] row;
	updateScope=writeScope;
	return dataCount;
}

// update line with data array and return number of data being updated
template <typename T>
size_t PortableAnyMap::UpdateLine(size_t line, T* data, size_t (*explicitCast)(T))
{
	Scope updateScope(scope.Width(),1,0,line);
	return UpdateScope(updateScope, data, explicitCast);
}

// write data to line and return number of data being written
template <typename T>
size_t PortableAnyMap::WriteLine(T* data, size_t (*explicitCast)(T))
{
	if (!CheckBeforeWriting(data))
		return 0;

	size_t dataCount=0;
	size_t val;

	// in case of ascii coding
	if (type<4)
	{
		// write line from data value by value to data array
		for (size_t i=0; i<rowSize; i++)
		{
			val=(explicitCast?explicitCast(*(data++)):(T)*(data++));
			if (type==PNM_TYPE_P1)
				fprintf(file, "%c ", (val?'1':'0'));
			else
				fprintf(file, "%u ", val);
			dataCount++;
		}
		fprintf(file,"\n");
	}
	else // in case of binary coding
	{
		// allocate memory tho read scope row at once
		unsigned char* row=new unsigned char[rowSize];
		unsigned char* rowPos=row;
		memset(row,0,rowSize);

		for (size_t i=0;i<scope.Width();i++)
		{
			// for pixmap loop through colors
			for (size_t j=0; j<(pixelSize<3?1:3); j++)
			{

				val=std::min((size_t)65535,(explicitCast?explicitCast(*(data++)):(size_t)*(data++)));
				if (pixelSize==0) // bitmap
				{
					int bitPos=7-i%8;
					if (val)
						*rowPos|=1<<bitPos;
					if (bitPos==0)
						rowPos++;
				}
				else	// pixmap or graymap
				{
					if (pixelSize%2==0)
					{
						*(rowPos++)=(unsigned char)(val>>8);
						*(rowPos++)=(unsigned char)(val);
					}
					else
						*(rowPos++)=std::min((size_t)255,val);
				}
				dataCount++;
			}
		}

		// write line
		fwrite(row,1,rowSize,file);
		delete [] row;
	}

	lineCount++;
	return dataCount;
}

// fill missing lines in map with data
inline bool PortableAnyMap::FillUp(size_t data_r, size_t data_g, size_t data_b)
{
	if (!CheckBeforeWriting(&data_r))
		return false;

	// adjust data_r_g_b values
	data_r=std::min(maxValue,data_r);
	data_g=(data_g==size_t(-1)?data_r:std::min(maxValue,data_g));
	data_b=(data_b==size_t(-1)?data_r:std::min(maxValue,data_b));

	// repeat data to fill line sized array
	size_t* data=new size_t[scope.Width()*(type%3?1:3)];
	size_t* pData=data;
	for (size_t i=0;i<scope.Width();i++)
	{
		*(pData++)=data_r;
		if (type%3==0)
		{
			*(pData++)=data_g;
			*(pData++)=data_b;
		}
	}

	// number of missing lines to file
	for (size_t i=lineCount; i<scope.Height(); i++)
		WriteLine(data);

	delete [] data;

	return true;
}
}
#endif // PGM_H_INCLUDED
